#!/usr/bin/env python

import sys
import os
import time
import subprocess as subp
import getopt
import argparse
import warnings
import math
from datetime import datetime

# List for available machines
# Update this list when adding new machine supports
available_machine_list = ['cori', 'edison','titan','summit', 'summitdev']
# List for GPU enabled machines
GPU_MACHINE_LIST = ['titan', 'summit', 'summitdev']

def write_jobscript(args):
    r'''write out GTC job script for all acceptable machines'''

    with open(args.job_filename, 'w+') as f:
        if (args.machine in ['cori', 'edison']):
            f.write('#!/bin/bash -l\n')
            f.write('#SBATCH --account={0}\n'.format(args.account))
            f.write('#SBATCH --partition={0}\n'.format(args.partition))
            f.write('#SBATCH --nodes={0}\n'.format(args.nodes))
            f.write('#SBATCH --time={0}\n'.format(args.time))
            f.write('#SBATCH --job-name={0}\n'.format(args.name))
            f.write('#SBATCH --workdir={0}\n'.format(args.workdir))
            if args.email is not None:
                f.write('#SBATCH --mail-user={0}\n'.format(args.email))
                f.write('#SBATCH --mail-type=ALL\n')
            if (args.machine == 'cori'):
                f.write('#SBATCH -C haswell\n')
        elif args.machine in ['titan']:
            f.write('#!/bin/sh \n')
            f.write('#PBS -A {0}\n'.format(args.account))
            f.write('#PBS -q {0}\n'.format(args.partition))
            f.write('#PBS -l nodes={0}\n'.format(args.nodes))
            f.write('#PBS -l walltime={0}\n'.format(args.time))
            f.write('#PBS -N {0}\n'.format(args.name))
            if args.email is not None:
                f.write('#PBS -M {0}\n'.format(args.email))
                f.write('#PBS -m abe\n') # send email for abort, begin and end
            f.write('\n')
        elif args.machine in ['summit', 'summitdev']:
            f.write('#!/bin/bash \n')
            f.write('#BSUB -P {0}\n'.format(args.account))
            f.write('#BSUB -W {0}\n'.format(args.time))
            f.write('#BSUB -alloc_flags gpudefault\n')
            f.write('#BSUB -alloc_flags smt{0}\n'.format(args.hypre_threads))
            f.write('#BSUB -o %J.out\n')
            f.write('#BSUB -e %J.err\n')
            f.write('#BSUB -J {0}\n'.format(args.name))
            f.write('#BSUB -nnodes {0}\n'.format(args.nodes))
            if args.email is not None:
                f.write('#BSUB -N\n')
            f.write('\n')
        else:
            raise ValueError('unsupported machine: {0}'.format(args.machine))

        # Debug on Titan needs to load module ddt
        if args.debug and (args.machine in ['titan']):
            f.write('source $MODULESHOME/init/bash\n')
            f.write('module load ddt\n')
        # OpenMP threads is by default always turned on
        f.write('export OMP_NUM_THREADS={0}\n'.format(args.threads))
        # GPU poisson parallelization control
        if args.machine in ['titan', 'summit', 'summitdev']:
            f.write('export GTC_RANKS_PER_SOLVE=1\n\n')
        # resource limit on summit and summitdev
        if args.machine in ['summit', 'summitdev']:
            f.write('ulimit -s 10240\n\n')
        # Change directory to working dir
        if args.machine in ['titan']:
            f.write('cd $PBS_O_WORKDIR\n\n')
        # Create necessary directories
        f.write('mkdir -p restart_dir1\n')
        f.write('mkdir -p restart_dir2\n')
        #f.write('mkdir -p restart_dir\n')
        f.write('mkdir -p phi_dir\n')
        f.write('mkdir -p trackp_dir\n\n')
        # Run the executable
        if (args.machine in ['cori', 'edison']):
            f.write('srun -n {0} -c $OMP_NUM_THREADS ./gtc\n'.format(args.nmpi))
        elif args.machine in ['titan']:
            if args.debug:
                f.write('ddt --connect ')
            f.write('aprun -n {0} -N {1} -d $OMP_NUM_THREADS ./gtc\n'\
                    .format(args.nmpi, args.mpi_per_node))
        elif args.machine in ['summit', 'summitdev']:
            _gpu = '-g1' if args.gpu else ''
            _run_command = 'jsrun -n {0} -r {1} {2} -a1'.format(args.nmpi,
                                                                args.mpi_per_node,
                                                                _gpu)
            _run_command += ' -c{0} -E OMP_NUM_THREADS={0} -b rs ./gtc\n'\
                    .format(args.threads)
            f.write(_run_command)
        f.write('\n')
        # Automatically increase irun by 1 in gtc.in
        f.write('irun_n=`sed -n \'s/^ *irun=\\([0-9]*\\).*/\\1/p\' gtc.in`'+'\n')
        f.write('(( irun_m = irun_n + 1 ))\n')
        f.write(r'sed -i "s/^\( *\)irun=${irun_n}/\1irun=${irun_m}/" gtc.in'+\
                '\n')
        # Automatically rename the gtc.out file by adding the irun number
        f.write('mv gtc.out gtc.out${irun_n}\n')

def test_job_finish(i, args):
    '''test if the number i queued job is completed

    :param int i: chained job number
    :param int waittime: persistent checking time in minutes
    '''
    gtcoutfile = './gtc.out{0}'.format(i)
    wait=0
    while wait < args.wait_time:
        if os.path.isfile(gtcoutfile):
            return True
        time.sleep(60)
        wait += 1
    print('Job number {0} exceeds '.format(i)+\
          'waiting time limit {0}min.'.format(args.wait_time)+\
          'Please check its status manually.')
    sys.exit(3)

def submit_job(i, args):
    '''submit the written job-script '''
    if args.machine in ['cori', 'edison']:
        submit_cmd = 'sbatch'
    elif args.machine in ['titan']:
        submit_cmd = 'qsub'
    elif args.machine in ['summit', 'summitdev']:
        submit_cmd = 'bsub'
    else:
        raise ValueError('Machine [{0}] Not Supported.'.format(args.machine)+\
        ' Write your own job-script writer.'+\
        ' If an existing machine job-script is supported,'+\
        ' add current machine to the available list in submit-GTC.py.'+\
        ' Available machines are {0}'.format(available_machine_list))
    subp.check_call([submit_cmd, args.job_filename])
    print('Job {0} submitted. {1}'.format(i,datetime.now()))

def parse_argument():
    """ define the argument parser, and parse the command line arguments
    """
    parser = argparse.ArgumentParser(description='submit GTC jobs')

    parser.add_argument('-e', dest='existing_jobfile', action='store_true',
                        help='If specified, will use the existing job script '+\
                             'file. Default: %(default)s')
    parser.add_argument('-m', '--machine', default='cori',
                        choices=available_machine_list,
                        help='Machine to be used. Default: %(default)s')
    parser.add_argument('-T', '--time', default='00:30:00',
                        help='requested wall time. Default: %(default)s')
    parser.add_argument('-p', '--partition', default='debug',
                        help='requested queue. Default: %(default)s')
    parser.add_argument('-N', '--nodes', default=1,type=int,
                        help='requested total node number. Default: %(default)s')
    parser.add_argument('-n', '--nmpi', default=None,type=int,
                        help='requested total MPI processes. '+\
                             'Default: total CPU devided by OpenMP threads.')
    parser.add_argument('-c', '--chain', default=1,type=int,
                        help='total job number to be chained. Default: %(default)s')
    parser.add_argument('--chain-start',dest='chain_start', default=0,type=int,
                        help='starting job/irun number of the chain. Default: %(default)s')
    parser.add_argument('-t','--threads', default=-1,type=int,
                        help='OpenMP threads number. Default: set by machine')
    parser.add_argument('-H','--hypre-threads', dest='hypre_threads', default=-1,type=int,
                        help='System core multi-threads number. Default: set by machine')
    parser.add_argument('--account', default='default',
                        help='Charging account code. Default: %(default)s')
    parser.add_argument('--email', default=None,
                        help='your email for receiving notifications. Default: %(default)s')
    parser.add_argument('--job-filename', dest='job_filename',
                        default='job.batch',
                        help='job batch script file name. Default: %(default)s')
    parser.add_argument('--wait-time', dest='wait_time', default=300, type=int,
                        help='Waiting time between chained jobs, in minutes. Default: %(default)s')
    parser.add_argument('--name', default='GTC',
                        help='Waiting time between chained jobs, in minutes. Default: %(default)s')
    parser.add_argument('--workdir', default='./',
                        help='Path to the gtc working directory. Default: %(default)s')
    parser.add_argument('--dry-run', dest='dry_run', action='store_true',
                        help='Generate the job script file without actually '+\
                             'submit it. Default: %(default)s')
    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='Prefix the DDT attachment command to the run.'+\
                             ' Default: %(default)s')
    parser.add_argument('--gpu', dest='gpu', action='store_true',
                        help='Use the GPU enabled settings'+\
                             ' Default: %(default)s')
    args = parser.parse_args()

    return args

def check_argument(args):
    """Check the validity and self-consistency of the args, fill in defaults"""

    if args.machine not in available_machine_list:
        raise ValueError('Machine [{0}] Not Supported.'.format(args.machine)+\
            ' Write your own job-script writer.'+\
            ' If an existing machine job-script is supported,'+\
            ' add current machine to the available list in submit-GTC.py.'+\
            ' Available machines are {0}'.format(available_machine_list))
    if (args.gpu is True) and args.machine not in GPU_MACHINE_LIST:
        warnings.warn('machine {0} is not in GPU_MACHINE_LIST, please check'\
                      .format(args.machine)+\
                      ' its GPU availability. GPU settings are turned off.')
        args.gpu = False
    if args.debug and args.machine in ['cori','edison']:
        warnings.warn('Debug on Cori or Edison doesn\'t need job scripts.'+\
                      ' The debug flag is reset to False.')
        args.debug = False


    # fill in the default values based on the chosen machine
    if args.account == 'default':
        if args.machine in ['edison', 'cori']:
            args.account = 'm808'
        elif args.machine in ['titan', 'summit', 'summitdev']:
            args.account = 'csc190gtc'
        else:
            raise ValueError('Default account not set for machine:' +\
                             ' {0}'.format(args.machine))

    if args.nmpi is None:
        # calculate the total MPI processes based on total nodes,
        # OMP_threads, and machine type
        if args.gpu is False:
            if args.machine == 'cori':
                ppn = 32
                if args.threads == -1:
                    args.threads = 2
                if args.hypre_threads == -1:
                    args.hypre_threads = 1
            elif args.machine == 'edison':
                ppn = 24
                if args.threads ==-1:
                    args.threads = 2
                if args.hypre_threads == -1:
                    args.hypre_threads = 1
            elif args.machine == 'titan':
                ppn = 16
                if args.threads ==-1:
                    args.threads = 2
                if args.hypre_threads == -1:
                    args.hypre_threads = 1
            elif args.machine == 'summit':
                ppn = 42
                if args.threads ==-1:
                    args.threads = 1
                if args.hypre_threads == -1:
                    args.hypre_threads = 4
            elif args.machine == 'summitdev':
                ppn = 20
                if args.threads ==-1:
                    args.threads = 1
                if args.hypre_threads == -1:
                    args.hypre_threads = 4
            else:
                raise ValueError('Wrong machine: {0}'.format(args.machine))

            # total allowed threads on one node
            ptpn = ppn * args.hypre_threads
            if ptpn % args.threads != 0:
                warnings.warn('OMP threads:{0}'.format(args.threads) +\
                              ' does not divide CPU-threads per node:{0}'\
                              .format(ptpn))
            args.mpi_per_node = int(ptpn/args.threads)
            args.nmpi = args.nodes*args.mpi_per_node
        else:
            if args.machine == 'titan':
                ppn = 1
                if args.threads ==-1:
                    args.threads = 1
                if args.hypre_threads == -1:
                    args.hypre_threads = 1
            elif args.machine == 'summitdev':
                ppn = 4
                if args.threads ==-1:
                    args.threads = 1
                if args.hypre_threads == -1:
                    args.hypre_threads = 1
            elif args.machine == 'summit':
                ppn = 6
                if args.threads ==-1:
                    args.threads = 7
                if args.hypre_threads == -1:
                    args.hypre_threads = 1
            args.mpi_per_node = ppn
            args.nmpi = args.nodes * args.mpi_per_node
    else:
        args.mpi_per_node = int(math.ceil(float(args.nmpi)/args.nodes))
    # desired physical core number is determined by the request OMP threads and
    # the hypre-threading setting
    args.core_per_mpi = int(math.ceil(float(args.threads)/args.hypre_threads))

def main():
    ''' create and submit GTC job-scripts corresponding to running environment
    '''
    # set the user warnings to be visible
    warnings.filterwarnings(action='always', category=UserWarning)

    # parse command-line arguments
    args = parse_argument()
    check_argument(args)

    # test if an existing job file is to be used
    if args.existing_jobfile:
        print('Using existing job file: {0}'.format(args.job_filename))
    else:
        #  if an existing job file is detected, ask the user whether overwrite it
        if os.path.isfile(args.job_filename):
            move = input('Existing job script file {0} detected.\n'.format(args.job_filename) +\
                             'Add -e flag to use existing job file by default.\n'+\
                             'How to proceed? ([Q]uit; [O]verwrite; [U]se existing file)\n')
            if move in ['o','O']:
                print('Existing job file overwritten. New file is used.')
                # write the job script
                write_jobscript(args)
            elif move in ['u', 'U']:
                print('Existing job file used.')
            else:
                print('Stop. No job submitted.')
                sys.exit(1)
        else:
            print('GTC submission parameters:')
            for key,value in vars(args).items():
                print('{0} : {1}'.format(key, value))

            # Double check the parameters before move on
            question = input('Hit Enter to continue(q to quit):')
            if question == 'q':
                print('User Quit.')
                sys.exit(1)

            write_jobscript(args)

    # submit the jobs
    if not args.dry_run:
        for i in range(args.chain_start, args.chain_start + args.chain):
            submit_job(i, args)
            # keep testing if the job is finished until wait_time is reached
            test_job_finish(i, args)


if __name__ == "__main__":
    main()
